/*
 * A VDR index file generator (C++)
 *
 * (C) 2003 Stefan Huelswitt <huels@iname.com>
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>

#include "file.h"
#include "pes.h"
#include "version.h"

#define MAX_FILE_NR 99

// Picture types:
#define I_FRAME    1
#define P_FRAME    2
#define B_FRAME    3

// --- cGenIndex ---------------------------------------------------------------

class cGenIndex : public cPES {
private:
  cFileName fileName;
  cIndex index;
  //
  int replayFile;
  int fileNo, fileOffset, packetOffset, packetNo;
  bool eof, scan;
  int state, skip;
  unsigned char buff[KILOBYTE(256)];
  int readNo, readOffset;
  int fileSize[MAX_FILE_NR+1];
  //
  bool NextFile(void);
protected:
  virtual int Output(const uchar *data, int len);
  virtual int Action1(uchar type, uchar *data, int len);
  virtual int Action2(uchar type, uchar *data, int len);
  virtual void Skipped(uchar *data, int len);
public:
  cGenIndex(void);
  ~cGenIndex();
  void Work(void);
  };

cGenIndex::cGenIndex(void)
{
  SetDefaultRule(prAct1);
  SetRuleR(0xE0,0xEF,prAct2);
}

cGenIndex::~cGenIndex()
{
}

bool cGenIndex::NextFile(void)
{
  if(replayFile>=0 && eof) {
    replayFile=fileName.NextFile();
    readNo=fileName.FileNumber();
    readOffset=0;
    fileSize[readNo]=fileName.FileSize();
    }
  eof=false;
  return replayFile>=0;
}

void cGenIndex::Work(void)
{
  eof=false;
  memset(fileSize,0,sizeof(fileSize));

  replayFile=fileName.Open();
  fileNo=readNo=fileName.FileNumber();
  fileSize[readNo]=fileName.FileSize();
  fileOffset=readOffset=0;

  if(replayFile>=0) {
    if(index.Open()) {
      while(NextFile()) {
        int count=read(replayFile,buff,sizeof(buff));
        if(count<0) {
          printf("read vdr: %s\n",strerror(errno));
          return;
          }
        else if(count==0) {
          if(fileSize[readNo]!=readOffset)
            printf("file %d read/size mismatch\n",readNo);
          eof=true;
          continue;
          }
        else {
          readOffset+=count;
          int used=Process(buff,count);
          if(count-used) printf("bummer, count!=0\n");
          }
        }
      Statistics();
      }
    }
  else printf("Failed to open 001.vdr. Is this a VDR recording dir?\n");
}

int cGenIndex::Output(const uchar *data, int len)
{
  if(scan) {
    for(int i=len ; i>0 ; i--) {
      uchar c=*data++;
      if(skip>0) { skip--; continue; }
      // searching for sequence 00 00 01 00 xx PIC
      switch(state) {
        case 0:
        case 1:
        case 3:
          if(c==0x00) state++; else state=0;
          break;
        case 2:
          if(c==0x01) state++; 
          else if(c!=0x00) state=0;
          break;
        case 4:
          state++;
          break;
        case 5:
          int type=(c>>3)&0x07;
          if(type>=I_FRAME && type<=B_FRAME) {
            index.Write(packetNo,packetOffset,type);
            //if(type==I_FRAME) printf("I_FRAME at %d\n",packetOffset);
            }
          else printf("unknown picture type %d at %d\n",type,packetOffset);
          scan=false; state=0; i=0;
          break;
        }
      }
    }

  Skipped(0,len);
  return len;
}

int cGenIndex::Action1(uchar type, uchar *data, int len)
{
  scan=false;
  return len;
}

int cGenIndex::Action2(uchar type, uchar *data, int len)
{
  if(SOP) {
    packetOffset=fileOffset; packetNo=fileNo;
    scan=true; skip=headerSize; state=0;
    }
  return len;
}

void cGenIndex::Skipped(uchar *data, int len)
{
  fileOffset+=len;
  if(fileOffset>=fileSize[fileNo]) {
    fileOffset-=fileSize[fileNo];
    fileNo++;
    }

  static int lastoff=0;
  if((fileOffset<lastoff || fileOffset>lastoff+KILOBYTE(256)) && fileSize[fileNo]) {
    printf("offset %d %d%%\r",fileOffset,(int)((long long)fileOffset*100/fileSize[fileNo])); fflush(stdout);
    lastoff=fileOffset;
    }
}

// --- MAIN --------------------------------------------------------------------

int main(int argc, char *argv[])
{
  printf("This is %s %s (%s), (C) 2003 Stefan Hlswitt\n"
         "Released under the GNU Public License\n",
         PRG_NAME,PRG_VERSION,__DATE__);

  cGenIndex gi;
  gi.Work();
  return 0;
}
