/*
 * genindex.c   generates index.vdr file from mpeg files written by VDR
 *
 * Compile:   gcc -o genindex -O2 -Wall -funroll-loops genindex.c
 *
 * Authors:   varies, including me (Werner Fink <werner@suse.de>)
 *
 * Changes:
 *
 *   Fri Mar 15, 2002: Move from random scanner to streamid scanner
 *   for getting the correct offsets of the I-Frames.  Be sure that
 *   no other multiplex than video ones will be scanned.  Increase both
 *   buffers for ringbuffer and stdio to lower I/O load and increase
 *   speed.   //werner
 *
 * Usage:
 *
 *   cd /video[<number>]/<Film>/<Title>/<date>.<time>.<prio>.<life>.rec/
 *   [mv -f index.vdr index.vdr.bak]
 *   [pathto/]genindex
 */
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#define BUFFER_SIZE   4096*128

#define SC_PICTURE    0x00

#define NO_PICTURE    0
#define I_FRAME     1
#define P_FRAME     2
#define B_FRAME     3

/* VDR supports only offset with sizeof(int) */
struct tIndex {int offset; unsigned char type; unsigned char number; short reserved; };

static FILE* fp;
static FILE* idx;
static long Foffset;
static char fname[20];
static unsigned char fbuffer[BUFFER_SIZE];

static unsigned char buf(long p)
{
  static unsigned char ringbuf[BUFFER_SIZE];

  if ((p >= Foffset) && (p < (Foffset + BUFFER_SIZE)))
    return ringbuf[p-Foffset];

  Foffset = p;
  if (fseek(fp, p, SEEK_SET) == 0)
  {
    size_t n = 0;

    memset(&ringbuf, 0, sizeof(ringbuf));
    n = fread(ringbuf, 1, sizeof(ringbuf), fp);

    if ((n < sizeof(ringbuf)) && ferror(fp))
    {
      fprintf(stderr, "Could not read from %s: %s\n", fname, strerror(errno));
      exit(1);
    }

    return ringbuf[p-Foffset];
  }
  return 0;
}

static int readfile(const unsigned char number)
{ 
  long filesize;
  long c = 0;

  sprintf(fname,"%03d.vdr", number);
  fp = fopen(fname, "r");
  if (!fp)
  {
    if (errno != ENOENT)
    {
      fprintf(stderr, "Could not open %s: %s\n", fname, strerror(errno));
      return -1;
    }
    putchar('\n');
    return 1;
  }
  setbuffer(fp, (char*)fbuffer, sizeof(fbuffer));
  if (fseek(fp, 0, SEEK_END) < 0)
  {
    fprintf(stderr, "Could not goto end of %s: %s\n", fname, strerror(errno));
    return -1;
  }
  if ((filesize = ftell(fp)) < 0)
  {
    fprintf(stderr, "Could not get size of %s: %s\n", fname, strerror(errno));
    return -1;
  }

  if (number > 1)
    putchar('\n');
  printf("Reading %s, %ld kB.\n", fname, (filesize >> 10));
  Foffset = filesize;

  rewind(fp); /* Even if done for p == 0 in buf() */

  while (c < filesize) {
		if (buf(c) == 0x00 && buf(c+1) == 0x00 && buf(c+2) == 0x01) {
			/* MPEG packet header */
			long id = buf(c+3);
			if (id == 0xBA) {
				/* MPEG1 PACK header */
				printf("MPEG1 PACK header\n");
				c+=12;
			} else { 
				long fs = ((buf(c+4) << 8) + buf(c+5)) + 6; /* packet size */
				if (id == 0xBB) {
					printf("MPEG1 System Header, fs=%d\n", fs);
				} else if (id >= 0xE0 && id <= 0xEF) {
					long hs = buf(c+8)+9;
					printf("MPEG video packet, hs=%d, fs=%d\n", hs, fs);
					long off = c+hs;
					for (; off < c+fs; ++off) {
						if (buf(off) == 0x00 && buf(off+1) == 0x00 && buf(off+2) == 0x01
								&& buf(off+3) == 0x00) {
							printf("Detected MPEG picture frame, type=%d\n", (buf(off+5)>>3) & 0x07);
							const unsigned char Ptype = (buf(off+5)>>3) & 0x07;
							if ((Ptype == I_FRAME) || (Ptype == P_FRAME) || (Ptype == B_FRAME))
							{
								/* VDR supports only offset with sizeof(int) */
								const struct tIndex i = {(int)c, Ptype, number, 0};

								if (fwrite(&i, 1, sizeof(i), idx) != sizeof(i))
								{
									fprintf(stderr, "Error writing index file: %s\n", strerror(errno));
									return -1;
								}
								if (Ptype == I_FRAME)
								{
									printf("I-Frame found at %ld kB\r", (c >> 10));
									fflush(stdout);
								}
		#if 1
								break;
		#endif
							}                 /* if Ptype == I_FRAME ... */
						}
					}
				} else if (id >= 0xA0 && id <= 0xAF) {
					printf("MPEG audio packet, fs=%d\n", fs);
				} else {
					printf("PES: id=0x%x, fs=%d\n", id, fs);
				}
				c+=fs;
			}
		} else {
			printf("PES error: 0x%x 0x%x 0x%x 0x%x\n", buf(c), buf(c+1), buf(c+2), 
					buf(c+3));
			c++;
			//exit(1);
		}
	}
#if 0			

    if (buf(c) == 0x00 && buf(c+1) == 0x00 && buf(c+2) == 0x01)
    {
      const long w = ((buf(c+4) << 8) | buf(c+5)) + 6;  /* width of frame */
      const long l = w + c;                 /* absolute length */
			long off;

      if (!((l > c) && (l <= filesize)))
        break;                      /* while c < filesize */

#if 1
      switch (buf(c+3))                   /* streamid */
      {
			case 0xE0:
			case 0xE1:
			case 0xE2:
			case 0xE3:
			case 0xE4:
			case 0xE5:
			case 0xE6:
			case 0xE7:
			case 0xE8:
			case 0xE9:
			case 0xEA:
			case 0xEB:
			case 0xEC:
			case 0xED:
			case 0xEE:
			case 0xEF:
			case 0xBA:
#endif
			off = (buf(c+8) + 9) + c;      /* absolute offset of frame contents */
			for (;off < l; off++)
			{
				if (buf(off) == 0x00 && buf(off+1) == 0x00 && buf(off+2) == 0x01 &&
					buf(off+3) == SC_PICTURE)
				{
					const unsigned char Ptype = (buf(off+5)>>3) & 0x07;
					if ((Ptype == I_FRAME) || (Ptype == P_FRAME) || (Ptype == B_FRAME))
					{
						/* VDR supports only offset with sizeof(int) */
						const struct tIndex i = {(int)c, Ptype, number, 0};

						if (fwrite(&i, 1, sizeof(i), idx) != sizeof(i))
						{
							fprintf(stderr, "Error writing index file: %s\n", strerror(errno));
							return -1;
						}
						if (Ptype == I_FRAME)
						{
							printf("I-Frame found at %ld kB\r", (c >> 10));
							fflush(stdout);
						}
#if 1
						break;
#endif
					}                 /* if Ptype == I_FRAME ... */
				}                   /* if buf(off) == 0x00 ... */
			}                     /* while off < l */
#if 1
			}
#endif
      c += w;                       /* Next frame */
    }
    else                          /* if buf(c) == 0x00 ... */
      c++;                        /* Broken file, search for begin */
  }                               /* while c < filesize */
#endif
  fclose(fp);
  return 0;
}

int main()
{
  int c, ret;
  if (!(idx = fopen("index.vdr", "w")))
  {
    fprintf(stderr, "Could not open index.vdr: %s\n", strerror(errno));
    return -1;
  }
  for (c=1; ((ret = readfile(c)) == 0); c++)
    __asm__ __volatile__("": : :"memory");
  fclose(idx);
  return (ret > 0 ? 0 : 1);
}
